def get_icon_file():
    import importlib
    from pathlib import Path

    module = importlib.import_module("manim_studio")
    module_path = Path(module.__file__).parent
    icon_folder = module_path / "assets"
    icon_file = icon_folder / "favicon.ico"
    return str(icon_file.absolute())



from PyQt6.QtWidgets import QVBoxLayout, QLabel, QSlider, QColorDialog, QComboBox, QLineEdit, QTextEdit, QCheckBox,\
    QPushButton, QScrollArea, QGroupBox, QDoubleSpinBox, QHBoxLayout
from PyQt6.QtGui import QColor
from PyQt6.QtCore import QThread, pyqtSignal, Qt
import socket
import numpy as np


class ClientControls(QScrollArea):
    add_slider_to_client = pyqtSignal(str, str, str, str, str)
    add_color_widget_to_client = pyqtSignal(str, int, int, int, int)
    add_dropdown_to_client = pyqtSignal(str, list, str)
    add_line_edit_to_client = pyqtSignal(str, str)
    add_text_editor_to_client = pyqtSignal(str, str)
    add_checkbox_to_client = pyqtSignal(str, bool)
    add_button_to_client = pyqtSignal(str)
    add_position_control_to_client = pyqtSignal(str, np.ndarray)
    set_control_value = pyqtSignal(str, str)
    close_signal = pyqtSignal()

    def __init__(self, main, s: socket.socket, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.main = main
        self.setWindowTitle("Manim Studio Client - Controls")
        self.layout_ = QVBoxLayout()
        self.controls = {}
        self.setLayout(self.layout_)
        self.label = QLabel(text="No controls defined")
        self.layout().addWidget(self.label)
        self.s = s
        self.add_slider_to_client.connect(
            self.add_slider_to_client_command)
        self.add_color_widget_to_client.connect(
            self.add_color_widget_to_client_command)
        self.add_dropdown_to_client.connect(
            self.add_dropdown_to_client_command)
        self.add_line_edit_to_client.connect(
            self.add_line_edit_to_client_command)
        self.add_text_editor_to_client.connect(
            self.add_text_editor_to_client_command)
        self.add_checkbox_to_client.connect(
            self.add_checkbox_to_client_command)
        self.add_button_to_client.connect(
            self.add_button_to_client_command)
        self.set_control_value.connect(
            self.set_control_value_command)
        self.add_position_control_to_client.connect(
            self.add_position_control_to_client_command)
        self.close_signal.connect(self.close)
        self.update_controls_thread = UpdateControlsThread(
            self.s,
            add_slider_to_client=self.add_slider_to_client,
            add_color_widget_to_client=self.add_color_widget_to_client,
            add_dropdown_to_client=self.add_dropdown_to_client,
            add_line_edit_to_client=self.add_line_edit_to_client,
            add_text_editor_to_client=self.add_text_editor_to_client,
            add_checkbox_to_client=self.add_checkbox_to_client,
            add_button_to_client=self.add_button_to_client,
            add_position_control_to_client=self.add_position_control_to_client,
            close_signal=self.close_signal,
            set_control_value=self.set_control_value)
        self.update_controls_thread.start()

    def set_control_value_command(self, name, value):
        if name in self.controls:
            control = self.controls[name]
            if isinstance(control, QSlider):
                control.setValue(int(value))
            elif isinstance(control, QColorDialog):
                r, g, b, a = value.split(",")
                control.setCurrentColor(
                    QColor(int(r), int(g), int(b), int(a)))
            elif isinstance(control, QComboBox):
                control.setCurrentText(value)
            elif isinstance(control, QLineEdit):
                control.setText(value)
            elif isinstance(control, QTextEdit):
                control.setText(value)
            elif isinstance(control, QCheckBox):
                control.setChecked(value == "True")
            elif isinstance(control, QGroupBox):
                x, y, z = value.split(",")
                control.x_.setValue(float(x))
                control.y_.setValue(float(y))
                control.z_.setValue(float(z))
            else:
                return
        else:
            return

    def add_slider_to_client_command(self, name, min_, max_, step, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        slider = QSlider()
        slider.setOrientation(Qt.Orientation.Horizontal)
        slider.setMinimum(int(min_))
        slider.setMaximum(int(max_))
        slider.setSingleStep(int(step))
        slider.setValue(int(default))
        slider.valueChanged.connect(
            lambda value: self.s.sendall(f"set_slider_value {name} {value}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(slider)
        self.controls[name] = slider

    def add_color_widget_to_client_command(self, name, r, g, b, a):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        color_dialog = QColorDialog()
        color_dialog.setCurrentColor(
            QColor(int(r), int(g), int(b), int(a)))
        color_dialog.currentColorChanged.connect(
            lambda color: self.s.sendall(f"set_color {name} {color.red()} {color.green()} {color.blue()} {color.alpha()}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(color_dialog)
        self.controls[name] = color_dialog

    def add_dropdown_to_client_command(self, name, options, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        dropdown = QComboBox()
        dropdown.addItems(options)
        dropdown.setCurrentText(default)
        dropdown.currentTextChanged.connect(
            lambda text: self.s.sendall(f"set_dropdown {name} {text}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(dropdown)
        self.controls[name] = dropdown

    def add_line_edit_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        line_edit = QLineEdit()
        line_edit.setText(default)
        line_edit.textChanged.connect(
            lambda text: self.s.sendall(f"set_line_edit {name} {text}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(line_edit)
        self.controls[name] = line_edit

    def add_text_editor_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        text_edit = QTextEdit()
        text_edit.setText(default)
        text_edit.textChanged.connect(
            lambda text: self.s.sendall(f"set_text_edit {name} {text}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(text_edit)
        self.controls[name] = text_edit

    def add_checkbox_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        checkbox = QCheckBox()
        checkbox.setChecked(default == "True")
        checkbox.stateChanged.connect(
            lambda: self.s.sendall(f"set_checkbox {name} {checkbox.isChecked()}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(checkbox)
        self.controls[name] = checkbox

    def add_button_to_client_command(self, name):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        button = QPushButton(text=name)
        button.clicked.connect(
            lambda: self.s.sendall(f"button_clicked {name}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(button)
        self.controls[name] = button

    def close(self):
        super().close()
        self.main.close()
        self.s.close()

    def add_position_control_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        widget = QGroupBox(name)
        widget.setLayout(QHBoxLayout())
        widget.x_ = QDoubleSpinBox()
        widget.x_.setValue(default[0])
        widget.y_ = QDoubleSpinBox()
        widget.y_.setValue(default[1])
        widget.z_ = QDoubleSpinBox()
        widget.z_.setValue(default[2])
        widget.x_.valueChanged.connect(
            lambda value: self.s.sendall(f"set_position_control_value_x {name} {value}".encode("utf-8")))
        widget.y_.valueChanged.connect(
            lambda value: self.s.sendall(f"set_position_control_value_y {name} {value}".encode("utf-8")))
        widget.z_.valueChanged.connect(
            lambda value: self.s.sendall(f"set_position_control_value_z {name} {value}".encode("utf-8")))
        widget.display_dot_checkbox = QCheckBox("Display Dot")
        widget.display_dot_checkbox.setChecked(False)
        widget.display_dot_checkbox.stateChanged.connect(
            lambda: self.s.sendall(f"set_position_control_display_dot {name} {widget.display_dot_checkbox.isChecked()}".encode("utf-8")))
        widget.layout().addWidget(widget.x_)
        widget.layout().addWidget(widget.y_)
        widget.layout().addWidget(widget.z_)
        widget.layout().addWidget(widget.display_dot_checkbox)
        self.layout().addWidget(widget)
        self.controls[name] = widget


class UpdateControlsThread(QThread):
    def __init__(self,
                 s: socket.socket,
                 add_slider_to_client: pyqtSignal,
                 add_color_widget_to_client: pyqtSignal,
                 add_dropdown_to_client: pyqtSignal,
                 add_line_edit_to_client: pyqtSignal,
                 add_text_editor_to_client: pyqtSignal,
                 add_checkbox_to_client: pyqtSignal,
                 add_button_to_client: pyqtSignal,
                 add_position_control_to_client: pyqtSignal,
                 close_signal: pyqtSignal,
                 set_control_value: pyqtSignal):
        super().__init__()
        self.s = s
        self.add_slider_to_client = add_slider_to_client
        self.add_color_widget_to_client = add_color_widget_to_client
        self.add_dropdown_to_client = add_dropdown_to_client
        self.add_line_edit_to_client = add_line_edit_to_client
        self.add_text_editor_to_client = add_text_editor_to_client
        self.add_checkbox_to_client = add_checkbox_to_client
        self.add_button_to_client = add_button_to_client
        self.add_position_control_to_client = add_position_control_to_client
        self.close_signal = close_signal
        self.set_control_value = set_control_value

    def run(self):
        self.s.sendall(b"get_controls")
        while True:
            try:
                msg = self.s.recv(1024)
            except ConnectionResetError:
                break
            command = msg.decode("utf-8")
            if command == "no_controls":
                continue
            if command.startswith("add_slider"):
                _, name, default, min_, max_, step = command.split(" ")
                self.add_slider_to_client.emit(
                    name, min_, max_, step, default)
            elif command.startswith("add_color_widget"):
                _, name, r, g, b, a = command.split(" ")
                self.add_color_widget_to_client.emit(
                    name, r, g, b, a)
            elif command.startswith("add_dropdown"):
                _, name, options, default = command.split(" ")
                options = options.split(",")
                self.add_dropdown_to_client.emit(
                    name, options, default)
            elif command.startswith("add_line_editor_widget"):
                _, name, *words = command.split(" ")
                self.add_line_edit_to_client.emit(
                    name, " ".join(words))
            elif command.startswith("add_text_editor_widget"):
                _, name, *words = command.split(" ")
                self.add_text_editor_to_client.emit(
                    name, " ".join(words))
            elif command.startswith("add_checkbox"):
                _, name, default = command.split(" ")
                self.add_checkbox_to_client.emit(
                    name, default == "True")
            elif command.startswith("add_button"):
                _, name = command.split(" ")
                self.add_button_to_client.emit(name)
            elif command.startswith("add_position_control"):
                _, name, *default = command.split(" ")
                x, y, z = default
                x = float(x)
                y = float(y)
                z = float(z)
                self.add_position_control_to_client.emit(
                    name, np.array([x, y, z]))
            elif command == "exit":
                self.close_signal.emit()
                break
            else:
                continue



from PyQt6.QtWidgets import QSlider

from .int_value_tracker import IntValueTracker


class Slider(QSlider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_tracker = IntValueTracker(0)
        self.valueChanged.connect(self.value_tracker.set_value)



from manim import *
from PyQt6.QtCore import QObject, pyqtSlot
from PyQt6.QtWidgets import QMessageBox, QDialog, QVBoxLayout, QLineEdit, QLabel, QPushButton, QFileDialog
from .communicate import Communicate
import dill as pickle

from .load_mobject import load_mobject


class CairoLiveRenderer(QObject, CairoRenderer):
    def __init__(self, communicate: Communicate, *args, **kwargs):
        QObject.__init__(self)
        CairoRenderer.__init__(self, *args, **kwargs)
        self.communicate = communicate

    def render(self, scene, time, moving_mobjects):
        super().render(scene, time, moving_mobjects)
        self.communicate.update_image.emit(self.get_frame())


class LiveScene(QObject, Scene):
    def __init__(self, communicate: Communicate, renderer=None, *args, **kwargs):
        config.disable_caching = True
        QObject.__init__(self)
        Scene.__init__(self, *args, **kwargs)
        self.communicate = communicate
        renderer = CairoLiveRenderer(communicate)
        QObject.__init__(self)
        Scene.__init__(self, renderer=renderer, *args, **kwargs)
        self.communicate.update_scene.connect(self.update_scene)
        self.communicate.end_scene.connect(self.end_scene)
        self.communicate.alert.connect(self.alert)
        self.communicate.save_mobject.connect(self.save_mobject)
        self.communicate.load_mobject.connect(self.load_mobject)
        self.current_code = None
        self.codes = []
        self.freeze = False
        self.playing = False

    def play(self, *args, **kwargs):
        if self.playing is True:
            alert = QMessageBox(
                text="You cannot play the scene while it is playing.")
            alert.setWindowTitle("Scene playing")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        self.playing = True
        super().play(*args, **kwargs)
        self.playing = False

    def set_control_value(self, name: str, value: str):
        self.communicate.set_control_value.emit(name, value)

    def construct(self):
        while True:
            while self.no_instruction():
                self.wait_until(lambda: not self.no_instruction(), 1)
            self.run_instruction()

    def add_checkbox_command(self, name: str, default_value: bool):
        self.communicate.add_checkbox_to_editor.emit(
            name, default_value)

    def add_text_editor_command(self, name: str, default_value: str):
        self.communicate.add_text_editor_to_editor.emit(
            name, default_value)

    def add_line_edit_command(self, name: str, default_value: str):
        self.communicate.add_line_edit_to_editor.emit(
            name, default_value)

    def add_dropdown_command(self, name: str, options: list[str], default_value: str):
        self.communicate.add_dropdown_to_editor.emit(
            name, options, default_value)

    def add_color_widget_command(self, name: str, default_value: np.ndarray):
        self.communicate.add_color_widget_to_editor.emit(
            name, default_value)

    def add_slider_command(self, name: str, default_value: str, min_value: str, max_value: str, step_value: str):
        self.communicate.add_slider_to_editor.emit(
            name, default_value, min_value, max_value, step_value)

    def add_button_command(self, name: str, callback: str):
        self.communicate.add_button_to_editor.emit(
            name, callback)

    def wait(self, *args, frozen_frame=False, **kwargs):
        super().wait(*args, frozen_frame=frozen_frame, **kwargs)

    def run_instruction(self):
        try:
            scope = globals()
            scope["self"] = self
            current_code = self.current_code
            self.current_code = None
            exec(current_code, scope)
        except EndSceneEarlyException:
            raise EndSceneEarlyException()
        except Exception as e:
            logger.info(
                f"Exception occured in live scene ({e.__class__.__name__}: {e})")
            self.communicate.alert.emit(e)
        else:
            self.codes.append(self.current_code)

    def add_position_control_command(self, name: str, default_value: np.ndarray):
        self.communicate.add_position_control_to_editor.emit(
            name, default_value)

    @pyqtSlot(str)
    def update_scene(self, code: str):
        if self.freeze and code != "raise EndSceneEarlyException()":
            alert = QMessageBox(
                text="You cannot update the scene while it is paused.")
            alert.setWindowTitle("Scene paused")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        self.current_code = code

    @pyqtSlot()
    def end_scene(self):
        self.current_code = "raise EndSceneEarlyException()"
        alert = QMessageBox(
            text="Your scene has ended. See the terminal for more information.")
        alert.setWindowTitle("Scene ended")
        alert.setIcon(QMessageBox.Icon.Information)
        alert.setStandardButtons(QMessageBox.StandardButton.Ok)
        alert.exec()

    @pyqtSlot(Exception)
    def alert(self, e: Exception):
        alert = QMessageBox(
            text="An error occured while processing your code.")
        alert.setWindowTitle("Error")
        alert.setIcon(QMessageBox.Icon.Warning)
        alert.setStandardButtons(QMessageBox.StandardButton.Ok)
        alert.setInformativeText(f"{e.__class__.__name__}: {e}")
        alert.exec()

    def pause_slide(self):
        self.freeze = True
        while self.freeze and self.no_instruction():
            self.wait_until(
                lambda: not self.freeze or not self.no_instruction(), 1)
        if not self.no_instruction():
            self.run_instruction()

    def no_instruction(self):
        return self.current_code is None

    def save_mobject(self):
        dialog = QDialog()
        dialog.setWindowTitle("Save mobject")
        dialog.layout_ = QVBoxLayout()
        dialog.name_edit = QLineEdit()
        dialog.name_edit.setPlaceholderText("Enter the name of the mobject")
        dialog.info_label = QLabel(
            text="The name of the mobject must be a valid attribute of the scene.\n"
            "For example, if you want to save the mobject 'self.circle',\n"
            "you must enter 'circle' in the text box below.")
        dialog.layout_.addWidget(dialog.name_edit)
        dialog.layout_.addWidget(dialog.info_label)
        dialog.ok_button = QPushButton("OK")
        dialog.ok_button.clicked.connect(dialog.accept)
        dialog.layout_.addWidget(dialog.ok_button)
        dialog.setLayout(dialog.layout_)
        dialog.exec()

        name = dialog.name_edit.text()
        if name == "":
            alert = QMessageBox(
                text="You must enter a name for the mobject.")
            alert.setWindowTitle("No name entered")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        mobject_to_save = getattr(self, name, None)
        if mobject_to_save is None:
            alert = QMessageBox(
                text="The name you entered is not a valid attribute of the scene.")
            alert.setWindowTitle("Invalid name")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        file_name = QFileDialog.getSaveFileName(
            caption="Save mobject",
            filter="Pickle file (*.pkl)"
        )
        if file_name[0]:
            with open(file_name[0], "wb") as f:
                pickle.dump(mobject_to_save, f)
        else:
            alert = QMessageBox(
                text="You must enter a file name.")
            alert.setWindowTitle("No file name entered")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        alert = QMessageBox(
            text="The mobject has been saved.")
        alert.setWindowTitle("Mobject saved")
        alert.setIcon(QMessageBox.Icon.Information)
        alert.setStandardButtons(QMessageBox.StandardButton.Ok)
        alert.exec()

    def load_mobject(self):
        file_name = QFileDialog.getOpenFileName(
            caption="Load mobject",
            filter="Pickle file (*.pkl)"
        )
        if not file_name[0]:
            alert = QMessageBox(
                text="You must enter a file name.")
            alert.setWindowTitle("No file name entered")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        dialog = QDialog()
        dialog.setWindowTitle("Load mobject")
        dialog.layout_ = QVBoxLayout()
        dialog.name_edit = QLineEdit()
        dialog.name_edit.setPlaceholderText("Enter the name of the mobject")
        dialog.info_label = QLabel(
            text="The name of the mobject will be an attribute of the scene.\n"
            "For example, if you want to load the mobject as 'self.circle',\n"
            "you must enter 'circle' in the text box below.")
        dialog.layout_.addWidget(dialog.name_edit)
        dialog.layout_.addWidget(dialog.info_label)
        dialog.ok_button = QPushButton("OK")
        dialog.ok_button.clicked.connect(dialog.accept)
        dialog.layout_.addWidget(dialog.ok_button)
        dialog.setLayout(dialog.layout_)
        dialog.exec()
        if dialog.name_edit.text() == "":
            alert = QMessageBox(
                text="You must enter a name for the mobject.")
            alert.setWindowTitle("No name entered")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        code = f"self.{dialog.name_edit.text()} = load_mobject('{file_name[0]}')"
        self.communicate.update_scene.emit(code)
        alert = QMessageBox(
            text="The mobject has been loaded.")
        alert.setWindowTitle("Mobject loaded")
        alert.setIcon(QMessageBox.Icon.Information)
        alert.setStandardButtons(QMessageBox.StandardButton.Ok)
        alert.exec()



from PyQt6.QtWidgets import QPushButton
from .communicate import Communicate


class Button(QPushButton):
    def __init__(self, callback: str, communicate: Communicate, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.callback = callback
        self.communicate = communicate
        self.clicked.connect(
            lambda: self.communicate.update_scene.emit(self.callback))



from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLineEdit, QPushButton


class ClientDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Manim Studio Client Dialog")
        self.host_edit = QLineEdit()
        self.host_edit.setPlaceholderText("Enter the host")
        self.port_edit = QLineEdit()
        self.port_edit.setPlaceholderText("Enter the port")
        self.username_edit = QLineEdit()
        self.username_edit.setPlaceholderText("Enter your username")
        self.password_edit = QLineEdit()
        self.password_edit.setPlaceholderText("Enter the password")
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.host_edit)
        self.layout.addWidget(self.port_edit)
        self.layout.addWidget(self.username_edit)
        self.layout.addWidget(self.password_edit)
        self.ok_button = QPushButton("OK")
        self.ok_button.clicked.connect(self.accept)
        self.layout.addWidget(self.ok_button)
        self.setLayout(self.layout)

    def get_server_info(self):
        host = self.host_edit.text()
        port = int(self.port_edit.text())
        username = self.username_edit.text()
        password = self.password_edit.text()
        return host, port, username, password




from PyQt6.QtWidgets import QDialog, QVBoxLayout, QCheckBox, QPushButton, QWidget


class ControlDialog(QDialog):
    def __init__(self, controls: dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle("Add controls to snippet")
        self.layout_ = QVBoxLayout()
        self.addControlsCheckbox = QCheckBox("Add controls")
        self.addControlsCheckbox.setChecked(False)
        self.add_controls = False
        self.controls_toggle = {k: False for k in controls.keys()}
        self.addControlsCheckbox.stateChanged.connect(self.toggle_add_controls)
        self.controls_toggle_widget = QWidget()
        self.controls_toggle_layout = QVBoxLayout()
        self.controls_toggle_widget.setLayout(self.controls_toggle_layout)
        for name in controls.keys():
            checkbox = QCheckBox(name)
            checkbox.setChecked(False)
            checkbox.stateChanged.connect(
                lambda state, name=name: self.toggle_control(name, state))
            self.controls_toggle_layout.addWidget(checkbox)
        self.controls_toggle_widget.hide()
        self.okButton = QPushButton("OK")
        self.okButton.clicked.connect(self.accept)
        self.layout_.addWidget(self.addControlsCheckbox)
        self.layout_.addWidget(self.okButton)
        self.layout_.addWidget(self.controls_toggle_widget)
        self.setLayout(self.layout_)

    def toggle_add_controls(self, state: bool):
        self.add_controls = bool(state)
        if self.add_controls:
            self.controls_toggle_widget.show()
        else:
            self.controls_toggle_widget.hide()

    def exec_and_return(self):
        self.exec()
        return self.add_controls, self.controls_toggle

    def toggle_control(self, name: str, state: bool):
        self.controls_toggle[name] = bool(state)



from PyQt6.QtWidgets import QApplication
from PyQt6.QtGui import QIcon
import sys

from .editor_widget import EditorWidget
from .preview_widget import PreviewWidget
from .live_scene import LiveScene
from .communicate import Communicate
from .ai_widget import AIWidget
from .get_icon_file import get_icon_file


def main(scene_type=LiveScene, server=False) -> None:
    app = QApplication(sys.argv)
    icon = QIcon(get_icon_file())
    app.setWindowIcon(icon)
    communicate = Communicate()
    scene = scene_type(communicate)
    editor_window = EditorWidget(communicate, scene, server)
    preview_window = PreviewWidget(communicate, scene)
    preview_window.showFullScreen()
    editor_window.show()
    if AIWidget is not None:
        ai_window = AIWidget(communicate)
        ai_window.show()
    if server:
        from .server import ManimStudioServer
        server_window = ManimStudioServer(communicate, editor_window)
        server_window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main(LiveScene)



from PyQt6.QtWidgets import QDoubleSpinBox, QGroupBox, QHBoxLayout, QCheckBox
from manim import Dot
from .live_scene import LiveScene
import numpy as np


class PositionControl(QGroupBox):
    def __init__(self, name: str, live_scene: LiveScene, default: np.ndarray = np.array([0, 0, 0])):
        self.name = name
        self.default = default
        self.live_scene = live_scene
        super().__init__(name)
        self.x_ = QDoubleSpinBox()
        self.y_ = QDoubleSpinBox()
        self.z_ = QDoubleSpinBox()
        self.display_dot_checkbox = QCheckBox("Display Dot")
        self.display_dot_checkbox.setChecked(False)
        self.display_dot_checkbox.stateChanged.connect(
            lambda: self.display_dot(self.display_dot_checkbox.isChecked()))
        self.dot = Dot(self.default).set_opacity(0)
        setattr(self.live_scene, self.name, self.dot)
        self.live_scene.add(getattr(self.live_scene, self.name))
        self.x_.setValue(default[0])
        self.y_.setValue(default[1])
        self.z_.setValue(default[2])
        self.x_.valueChanged.connect(self.update_x)
        self.y_.valueChanged.connect(self.update_y)
        self.z_.valueChanged.connect(self.update_z)
        self.setLayout(QHBoxLayout())
        self.layout().addWidget(self.x_)
        self.layout().addWidget(self.y_)
        self.layout().addWidget(self.z_)
        self.layout().addWidget(self.display_dot_checkbox)

    def update_x(self, value: int):
        self.default[0] = value
        self.dot.set_x(value)

    def update_y(self, value: int):
        self.default[1] = value
        self.dot.set_y(value)

    def update_z(self, value: int):
        self.default[2] = value
        self.dot.set_z(value)

    def display_dot(self, display: bool):
        if display:
            self.dot.set_opacity(1)
        else:
            self.dot.set_opacity(0)



from manim import logger, config, SVGMobject, VMobject, color_to_rgb
from colour import Color
from manim.utils.tex_file_writing import tex_hash
import inspect
import importlib
from pathlib import Path
import subprocess
import re


def create_dir_if_not_exists(dir_path: Path):
    if not dir_path.exists():
        dir_path.mkdir(parents=True)


def check_nodejs_installation():
    try:
        subprocess.run(["node", "--version"], check=True)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Node.js is not installed. Please install node and try again.")
    except subprocess.CalledProcessError:
        raise RuntimeError(
            "Node.js is not installed correctly. Please install node and try again.")


def get_cli_path():
    module = importlib.import_module(
        "manim_studio.mathjax_mobject.mathjax_mobject")
    folder = Path(inspect.getfile(module)).parent
    return (folder / "mathjax-cli.js").absolute()


def install_dependencies_if_not_installed():
    logger.info("Checking if manim-mathjax is installed...")
    logger.info("Installing manim-mathjax...")
    try:
        subprocess.run(
            ["npm", "install", get_cli_path().parent], check=True, shell=True)
    except FileNotFoundError:
        raise FileNotFoundError(
            "NPM is not installed. Please install npm and try again.")
    except subprocess.CalledProcessError:
        raise RuntimeError(
            "An error occurred while installing manim-mathjax. Please try again.")
    logger.info("manim-mathjax is installed.")


def render_mathjax(mathjax_str: str, check_installation=True):
    """Render a MathJax string as a mobject.

    Solution from https://github.com/manim-kindergarten/ManimGL-MathJax/blob/main/manimgl_mathjax/mathjax.py

    Parameters
    ----------
    mathjax_str : str
        A MathJax string.
    **kwargs
        Keyword arguments to pass to the MathTex mobject.

    Returns
    -------
    MathTex
        A MathTex mobject.
    """
    if check_installation:
        check_nodejs_installation()
        install_dependencies_if_not_installed()
    cli_path = get_cli_path()
    p = subprocess.Popen(["node", str(cli_path)],
                         stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    out, err = p.communicate(input=mathjax_str.encode("utf-8"))
    svg_str = out.decode("utf-8")

    error_match_obj = re.search(
        r"(?<=data\-mjx\-error\=\")(.*?)(?=\")", svg_str)
    if error_match_obj is not None:
        logger.error(
            "LaTeX Error!  Not a worry, it happens to the best of us.")
        logger.info(f"The error could be: `{error_match_obj.group()}`")
        raise ValueError("LaTeX Error")
    tex_dir = config.get_dir("tex_dir")
    tex_file = tex_dir / f"{tex_hash(mathjax_str)}.svg"
    create_dir_if_not_exists(config.get_dir("media_dir"))
    create_dir_if_not_exists(tex_dir)
    with open(tex_file, "wb") as f:
        f.write(out)
    return tex_file


class MathJax(SVGMobject):
    first_time = True

    def __init__(self, expr: str, font_size: int = 48, color=None, opacity=None, stroke_width=None, **kwargs):
        self.expr = expr
        self.font_size = font_size
        if color is None:
            color = VMobject().color
        if opacity is None:
            opacity = 1.0
        if stroke_width is None:
            stroke_width = 0.0
        self.colored_expr = self.get_colored_expr(self.expr, color)
        super().__init__(render_mathjax(self.colored_expr, self.__class__.first_time), **kwargs)
        self.__class__.first_time = False
        self.set_color(color, family=False)
        self.set_opacity(opacity)
        self.set_stroke(width=stroke_width)
        self.scale(0.2 * self.font_size / 48)

    def get_colored_expr(self, expr: str, color: Color):
        return f"\\textcolor[rgb]{{{','.join(map(str, color_to_rgb(color)))}}}{{{expr}}}"



from manim_studio.server.server import Server
from PyQt6.QtWidgets import QWidget, QInputDialog, QLabel, QVBoxLayout, QPushButton, QLineEdit, QDialog
from pyngrok import ngrok


class ManimStudioServer(QWidget):
    def __init__(self, communicate, editor):
        super().__init__()
        self.setWindowTitle("Manim Studio Server")
        self.communicate = communicate
        password_dialog = QInputDialog()
        password_dialog.setLabelText("Create a password for the server: ")
        password_dialog.setWindowTitle("Manim Studio Server")
        password_dialog.setModal(True)
        password_dialog.exec()
        self.password = password_dialog.textValue()
        self.server = Server("", 5555, self.password, self.communicate, editor)
        self.server.start()
        self.label = QLabel()
        self.label.setText(
            f"Server started at {self.server.host}:{self.server.port}")
        self.ask_label = QLabel()
        self.ask_label.setText(
            "If you want to work collaboratively, you need to have installed ngrok.\nIf you click the button below, ngrok will be started and you will be able to share the link with your mates.")
        self.ask_label.setWordWrap(True)
        self.ask_label.setMargin(10)
        self.ask_label.setIndent(10)
        self.public_url = None
        self.ask_label.setStyleSheet("background-color: #f0f0f0;")
        self.ngrok_button = QPushButton("Start ngrok")
        self.ngrok_button.clicked.connect(self.start_ngrok)
        self.users_button = QPushButton("Users")
        self.users_button.clicked.connect(self.show_users)
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.ask_label)
        self.layout.addWidget(self.ngrok_button)
        self.layout.addWidget(self.users_button)
        self.setLayout(self.layout)
        editor.manim_studio_server = self.server

    def closeEvent(self, event):
        self.server.close()
        if self.public_url is not None:
            self.stop_ngrok()
        super().closeEvent(event)

    def show_users(self):
        dialog = QDialog()
        layout = QVBoxLayout()
        dialog.setLayout(layout)
        users_dict = self.server.users
        for username in users_dict.keys():
            label = QLabel()
            label.setText(
                f"{username}")
            disconnect_user_button = QPushButton(f"Disconnect {username}")
            disconnect_user_button.clicked.connect(
                lambda: self.server.disconnect_user(username))
            disconnect_user_button.clicked.connect(
                lambda: label.hide())
            disconnect_user_button.clicked.connect(
                lambda: disconnect_user_button.hide())
            dialog.layout().addWidget(label)
            dialog.layout().addWidget(disconnect_user_button)
        update_button = QPushButton("Update")
        dialog.layout().addWidget(update_button)

        def update_button_callback():
            dialog.accept()
            self.show_users()

        update_button.clicked.connect(update_button_callback)
        dialog.setWindowTitle("Users")
        dialog.exec()

    def start_ngrok(self):
        tunnel = ngrok.connect(5555, "tcp")
        self.ngrok_button.setText("ngrok started")
        self.ngrok_button.setEnabled(False)
        self.link_label = QLabel()
        self.link_label.setText(
            "ngrok started. You can share the host and port below with your mates.")
        self.link_label.setWordWrap(True)
        self.link_label.setMargin(10)
        self.link_label.setIndent(10)
        self.link_label.setStyleSheet("background-color: #f0f0f0;")
        self.layout.addWidget(self.link_label)
        self.link = QLineEdit()
        self.public_url = tunnel.public_url
        host, port = tunnel.public_url[6:].split(":")
        self.link.setText(f"HOST={host}; PORT={port}")
        self.link.setReadOnly(True)
        self.layout.addWidget(self.link)

    def stop_ngrok(self):
        ngrok.disconnect(self.public_url)
        ngrok.kill()



from manim import ValueTracker, Mobject


class StringValueTracker(ValueTracker):
    def __init__(self, value="", **kwargs):
        Mobject.__init__(self, **kwargs)
        self.set_value(value)

    def get_value(self) -> str:
        return str(self.value)

    def set_value(self, value: str):
        self.value = str(value)



from PyQt6.QtWidgets import QComboBox

from .string_value_tracker import StringValueTracker


class DropdownWidget(QComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_tracker = StringValueTracker("")
        self.currentTextChanged.connect(self.value_tracker.set_value)



from manim import ValueTracker


class IntValueTracker(ValueTracker):
    def get_value(self) -> int:
        return int(self.points[0, 0])

    def set_value(self, value: int):
        self.points[0, 0] = int(value)
        return self



from .main import main as run_manim_studio
from .live_scene import LiveScene
from .load_mobject import load_mobject
from .load_snippet import load_snippet
from .prerender_latex import prerender_latex
from .int_value_tracker import IntValueTracker
from .color_value_tracker import ColorValueTracker
from .string_value_tracker import StringValueTracker
from .boolean_value_tracker import BooleanValueTracker
from .mathjax_mobject.mathjax_mobject import MathJax



from pathlib import Path
from manim import logger
import subprocess


BUILTINS_REPO = "https://github.com/MathItYT/manim-studio-snippets.git"


def main():
    if not Path("builtin_snippets").exists():
        subprocess.run(
            [
                "git",
                "clone",
                BUILTINS_REPO,
                "builtin_snippets",
            ],
            check=True,
        )
        (Path("builtin_snippets") / ".git").unlink()
    else:
        logger.error(
            "If you want to initialize built-in snippets, please delete the builtins_snippets folder and try again!"
        )
        exit(1)



from PyQt6.QtWidgets import QWidget, QTextEdit, QPushButton, QVBoxLayout, QFileDialog, \
    QMenuBar, QMessageBox, QDialog, QLineEdit, QLabel, QCheckBox, QGridLayout, QScrollArea, \
    QDoubleSpinBox, QStatusBar
from PyQt6.QtGui import QAction, QIntValidator, QColor
from PyQt6.QtCore import pyqtSlot, Qt
import numpy as np
from pathlib import Path
import pickle
import socket

from .communicate import Communicate
from .live_scene import LiveScene
from .slider import Slider
from .color_widget import ColorWidget
from .dropdown_widget import DropdownWidget
from .line_editor_widget import LineEditorWidget
from .text_editor_widget import TextEditorWidget
from .checkbox_widget import CheckboxWidget
from .control_dialog import ControlDialog
from .button import Button
from .position_control import PositionControl


class EditorWidget(QWidget):
    def __init__(self, communicate: Communicate, scene: LiveScene, server: bool, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.communicate = communicate
        self.manim_studio_server = None
        self.server = server
        self.setWindowTitle("Manim Studio - Editor")
        self.scene = scene
        self.controls = dict()
        self.ready_to_save = True

        self.code_cell_edit = QTextEdit()
        self.code_cell_edit.setPlaceholderText("Enter your code here")
        self.code_cell_edit.setGeometry(0, 0, 1920, 250)

        self.send_button = QPushButton("Send code (Ctrl+Return)")
        self.send_button.setGeometry(0, 0, 100, 50)
        self.send_button.setShortcut("Ctrl+Return")
        self.send_button.clicked.connect(self.send_code)
        self.end_button = QPushButton("End scene without saving (Ctrl+Q)")
        self.end_button.setGeometry(0, 0, 100, 50)
        self.end_button.setShortcut("Ctrl+Q")
        self.end_button.clicked.connect(self.end_scene)
        self.end_and_save_button = QPushButton("End scene and save (Ctrl+S)")
        self.end_and_save_button.setShortcut("Ctrl+S")
        self.end_and_save_button.setGeometry(0, 0, 100, 50)
        self.end_and_save_button.clicked.connect(self.end_scene_saving)
        self.save_snip_button = QPushButton("Save snippet (Ctrl+Shift+S)")
        self.save_snip_button.setShortcut("Ctrl+Shift+S")
        self.save_snip_button.setGeometry(0, 0, 100, 50)
        self.save_snip_button.clicked.connect(self.save_snippet)
        self.save_snip_and_run_button = QPushButton(
            "Save snippet and run (Ctrl+Shift+Return)")
        self.save_snip_and_run_button.setShortcut("Ctrl+Shift+Return")
        self.save_snip_and_run_button.setGeometry(0, 0, 100, 50)
        self.save_snip_and_run_button.clicked.connect(
            self.save_snippet_and_run)
        self.communicate.save_snippet.connect(self.save_snippet_command)
        self.next_slide_button = QPushButton("Next slide (Ctrl+N)")
        self.next_slide_button.setShortcut("Ctrl+N")
        self.next_slide_button.setGeometry(0, 0, 100, 50)
        self.communicate.next_slide.connect(self.next_slide)
        self.next_slide_button.clicked.connect(
            self.communicate.next_slide.emit)

        self.menu_bar = QMenuBar()
        self.file_menu = self.menu_bar.addMenu("File")
        self.open_snip_action = QAction("Open snippet", self)
        self.open_snip_action.setShortcut("Ctrl+O")
        self.open_snip_action.triggered.connect(self.open_snippet)
        self.file_menu.addAction(self.open_snip_action)
        self.open_snip_and_run_action = QAction(
            "Open snippet and run", self)
        self.open_snip_and_run_action.setShortcut("Ctrl+Shift+O")
        self.open_snip_and_run_action.triggered.connect(
            self.open_snippet_and_run)
        self.file_menu.addAction(self.open_snip_and_run_action)
        self.edit_menu = self.menu_bar.addMenu("Edit")
        self.add_slider_action = QAction("Add slider", self)
        self.add_slider_action.setShortcut("Ctrl+Shift+1")
        self.add_slider_action.triggered.connect(
            self.add_slider)
        self.edit_menu.addAction(self.add_slider_action)
        self.add_color_widget_action = QAction("Add color widget", self)
        self.add_color_widget_action.setShortcut("Ctrl+Shift+2")
        self.add_color_widget_action.triggered.connect(
            self.add_color_widget)
        self.edit_menu.addAction(self.add_color_widget_action)
        self.add_dropdown_action = QAction("Add dropdown", self)
        self.add_dropdown_action.setShortcut("Ctrl+Shift+3")
        self.add_dropdown_action.triggered.connect(
            self.add_dropdown)
        self.edit_menu.addAction(self.add_dropdown_action)
        self.add_line_editor_widget_action = QAction(
            "Add line editor widget", self)
        self.add_line_editor_widget_action.setShortcut("Ctrl+Shift+4")
        self.add_line_editor_widget_action.triggered.connect(
            self.add_line_editor_widget)
        self.edit_menu.addAction(self.add_line_editor_widget_action)
        self.add_text_editor_widget_action = QAction(
            "Add text editor widget", self)
        self.add_text_editor_widget_action.setShortcut("Ctrl+Shift+5")
        self.add_text_editor_widget_action.triggered.connect(
            self.add_text_editor_widget)
        self.edit_menu.addAction(self.add_text_editor_widget_action)
        self.add_checkbox_widget_action = QAction(
            "Add checkbox widget", self)
        self.add_checkbox_widget_action.setShortcut("Ctrl+Shift+6")
        self.add_checkbox_widget_action.triggered.connect(
            self.add_checkbox_widget)
        self.edit_menu.addAction(self.add_checkbox_widget_action)
        self.add_button_widget_action = QAction(
            "Add button widget", self)
        self.add_button_widget_action.setShortcut("Ctrl+Shift+7")
        self.add_button_widget_action.triggered.connect(
            self.add_button_widget)
        self.edit_menu.addAction(self.add_button_widget_action)
        self.add_position_control_action = QAction(
            "Add position control", self)
        self.add_position_control_action.setShortcut("Ctrl+Shift+8")
        self.add_position_control_action.triggered.connect(
            self.add_position_control)
        self.edit_menu.addAction(self.add_position_control_action)
        self.save_mobject_action = QAction("Save mobject", self)
        self.save_mobject_action.setShortcut("Ctrl+Shift+9")
        self.save_mobject_action.triggered.connect(self.save_mobject)
        self.edit_menu.addAction(self.save_mobject_action)
        self.load_mobject_action = QAction("Load mobject", self)
        self.load_mobject_action.setShortcut("Ctrl+Shift+0")
        self.load_mobject_action.triggered.connect(self.load_mobject)
        self.edit_menu.addAction(self.load_mobject_action)

        self.layout_ = QVBoxLayout()
        self.layout_.addWidget(self.menu_bar)
        self.layout_.addWidget(self.code_cell_edit)
        self.layout_.addWidget(self.send_button)
        self.layout_.addWidget(self.end_button)
        self.layout_.addWidget(self.end_and_save_button)
        self.layout_.addWidget(self.save_snip_button)
        self.layout_.addWidget(self.save_snip_and_run_button)
        self.layout_.addWidget(self.next_slide_button)
        self.controls_widget = QWidget()
        self.controls_scroll = QScrollArea()
        self.controls_scroll.setWidget(self.controls_widget)
        self.controls_scroll.setWidgetResizable(True)
        self.no_controls = QLabel(text="No controls defined")
        self.controls_scroll.setWindowTitle("Manim Studio - Controls")
        self.controls_layout = QGridLayout()
        self.controls_layout.addWidget(self.no_controls)
        self.controls_widget.setLayout(self.controls_layout)
        self.status_bar = QStatusBar()
        self.layout_.addWidget(self.status_bar)
        self.communicate.add_slider_to_editor.connect(
            self.add_slider_to_editor)
        self.communicate.add_color_widget_to_editor.connect(
            self.add_color_widget_to_editor)
        self.communicate.add_dropdown_to_editor.connect(
            self.add_dropdown_to_editor)
        self.communicate.add_line_edit_to_editor.connect(
            self.add_line_editor_widget_to_editor)
        self.communicate.add_text_editor_to_editor.connect(
            self.add_text_editor_to_editor)
        self.communicate.add_checkbox_to_editor.connect(
            self.add_checkbox_to_editor)
        self.communicate.set_control_value.connect(self.set_control_value)
        self.communicate.add_button_to_editor.connect(
            self.add_button_to_editor)
        self.communicate.add_controls_to_client.connect(
            self.add_controls_to_client)
        self.communicate.press_button.connect(self.press_button)
        self.communicate.add_controls_to_clients.connect(
            self.add_controls_to_clients)
        self.communicate.add_position_control_to_editor.connect(
            self.add_position_control_to_editor)
        self.communicate.set_position_control_x.connect(
            lambda name, value: self.controls[name].x_.setValue(float(value)))
        self.communicate.set_position_control_y.connect(
            lambda name, value: self.controls[name].y_.setValue(float(value)))
        self.communicate.set_position_control_z.connect(
            lambda name, value: self.controls[name].z_.setValue(float(value)))
        self.communicate.set_position_control_display_dot.connect(
            lambda name, value: self.controls[name].display_dot_checkbox.setChecked(value == "True"))
        self.communicate.show_in_status_bar.connect(
            self.status_bar.showMessage)
        self.setLayout(self.layout_)

    def show(self):
        super().show()
        self.controls_scroll.show()

    def add_position_control(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add position control")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Position control name")
        x_label = QLabel("X", dialog)
        x_edit = QDoubleSpinBox(dialog)
        x_edit.setValue(0)
        y_label = QLabel("Y", dialog)
        y_edit = QDoubleSpinBox(dialog)
        y_edit.setValue(0)
        z_label = QLabel("Z", dialog)
        z_edit = QDoubleSpinBox(dialog)
        z_edit.setValue(0)
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_position_control_command(
            text_edit.text(), np.array([x_edit.value(), y_edit.value(), z_edit.value()])))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(x_label)
        layout.addWidget(x_edit)
        layout.addWidget(y_label)
        layout.addWidget(y_edit)
        layout.addWidget(z_label)
        layout.addWidget(z_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    def add_position_control_to_editor(self, name: str, default_value: np.ndarray):
        position_control = PositionControl(name, self.scene, default_value)
        self.controls[name] = position_control
        self.controls_layout.addWidget(position_control)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: position_control})

    def add_controls_to_client(self, s: socket.socket, controls: dict):
        if len(controls) == 0:
            s.sendall(b"no_controls")
            return
        for name, control in controls.items():
            if isinstance(control, Slider):
                s.sendall(f"add_slider {name} {control.value()} {control.minimum()} {control.maximum()} {control.singleStep()}".encode(
                    "utf-8"))
            elif isinstance(control, ColorWidget):
                s.sendall(f"add_color_widget {name} {control.getColor().red()} {control.getColor().green()} {control.getColor().blue()} {control.getColor().alpha()}".encode(
                    "utf-8"))
            elif isinstance(control, DropdownWidget):
                s.sendall(f"add_dropdown {name} {','.join([control.itemText(i) for i in range(control.count())])} {control.currentText()}".encode(
                    "utf-8"))
            elif isinstance(control, LineEditorWidget):
                s.sendall(f"add_line_editor_widget {name} {control.text()}".encode(
                    "utf-8"))
            elif isinstance(control, TextEditorWidget):
                s.sendall(f"add_text_editor_widget {name} {control.toPlainText()}".encode(
                    "utf-8"))
            elif isinstance(control, CheckboxWidget):
                s.sendall(f"add_checkbox {name} {control.isChecked()}".encode(
                    "utf-8"))
            elif isinstance(control, Button):
                s.sendall(f"add_button {name}".encode(
                    "utf-8"))
            elif isinstance(control, PositionControl):
                s.sendall(f"add_position_control {name} {control.x_.value()} {control.y_.value()} {control.z_.value()}".encode(
                    "utf-8"))

    def add_controls_to_clients(self, clients: list[socket.socket], controls: dict):
        for client in clients:
            self.add_controls_to_client(client, controls)

    def press_button(self, name: str):
        self.communicate.update_scene.emit(self.controls[name].callback)

    def add_button_widget(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add button widget")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Button widget name")
        callback_edit = QTextEdit(dialog)
        callback_edit.setPlaceholderText("Callback")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_button_command(
            text_edit.text(), callback_edit.toPlainText()))

        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(callback_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    def add_button_to_editor(self, name: str, callback: str):
        button = Button(callback, self.communicate, text=name)
        self.controls[name] = button
        self.controls_layout.addWidget(button)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: button})

    @pyqtSlot(str, str)
    def set_control_value(self, name: str, value: str):
        control = self.controls.get(name)
        if control is None:
            alert = QMessageBox()
            alert.setText(f"Control {name} not found")
            alert.exec()
            return
        if isinstance(control, Slider):
            control.setValue(int(value))
        elif isinstance(control, ColorWidget):
            value = value.split(",")
            control.setCurrentColor(QColor(int(value[0]), int(
                value[1]), int(value[2]), int(value[3])))
        elif isinstance(control, DropdownWidget):
            control.setCurrentText(value)
        elif isinstance(control, LineEditorWidget):
            control.setText(value)
        elif isinstance(control, TextEditorWidget):
            control.setText(value)
        elif isinstance(control, CheckboxWidget):
            control.setChecked(value == "True")
        elif isinstance(control, Button):
            control.clicked.disconnect()
            control.clicked.connect(
                lambda: self.communicate.update_scene.emit(value))
        elif isinstance(control, PositionControl):
            value = value.split(",")
            control.x_.setValue(float(value[0]))
            control.y_.setValue(float(value[1]))
            control.z_.setValue(float(value[2]))

    def save_mobject(self):
        self.communicate.save_mobject.emit()

    def load_mobject(self):
        self.communicate.load_mobject.emit()

    def add_text_editor_widget(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add text editor widget")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Text editor widget name")
        default_value_edit = QCheckBox(dialog)
        default_value_edit.setText("Is default value True?")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_text_editor_command(
            text_edit.text(), default_value_edit.text()))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(default_value_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    @pyqtSlot(str, bool)
    def add_checkbox_to_editor(self, name: str, default_value: bool):
        label = QLabel(text=name)
        checkbox = CheckboxWidget(name)
        checkbox.setChecked(default_value)
        self.controls[name] = checkbox
        setattr(self.scene, name, checkbox.value_tracker)
        self.controls_layout.addWidget(label)
        self.controls_layout.addWidget(checkbox)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: checkbox})

    @pyqtSlot(str, str)
    def add_text_editor_to_editor(self, name: str, default_value: str):
        label = QLabel(text=name)
        text_editor_widget = TextEditorWidget(name)
        text_editor_widget.setText(default_value)
        self.controls[name] = text_editor_widget
        setattr(self.scene, name, text_editor_widget.value_tracker)
        self.controls_layout.addWidget(label)
        self.controls_layout.addWidget(text_editor_widget)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: text_editor_widget})

    def add_checkbox_widget(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add checkbox widget")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Checkbox widget name")
        default_value_edit = QCheckBox(dialog)
        default_value_edit.setText("Is default value True?")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_checkbox_command(
            text_edit.text(), default_value_edit.isChecked()))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(default_value_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    def add_dropdown(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add dropdown")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Dropdown name")
        options_edit = QLineEdit(dialog)
        options_edit.setPlaceholderText("Dropdown options")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        default_value_edit = QLineEdit(dialog)
        default_value_edit.setPlaceholderText("Default value")
        ok_button.clicked.connect(lambda: self.scene.add_dropdown_command(
            text_edit.text(), options_edit.text().split(","), default_value_edit.text()))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(options_edit)
        layout.addWidget(default_value_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    def add_line_editor_widget(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add line editor widget")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Line editor widget name")
        default_value_edit = QLineEdit(dialog)
        default_value_edit.setPlaceholderText("Default value")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_line_edit_command(
            text_edit.text(), default_value_edit.text()))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(default_value_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    @pyqtSlot(str, str)
    def add_line_editor_widget_to_editor(self, name: str, default_value: str):
        label = QLabel(text=name)
        line_editor_widget = LineEditorWidget(name)
        line_editor_widget.setText(default_value)
        self.controls[name] = line_editor_widget
        setattr(self.scene, name, line_editor_widget.value_tracker)
        self.controls_layout.addWidget(label)
        self.controls_layout.addWidget(line_editor_widget)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: line_editor_widget})

    def send_code(self):
        self.communicate.update_scene.emit(self.code_cell_edit.toPlainText())
        self.code_cell_edit.clear()

    def add_slider(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add slider")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Slider name")
        default_value_edit = QLineEdit(dialog)
        default_value_edit.setValidator(QIntValidator(
            -2147483648, 2147483647))
        default_value_edit.setPlaceholderText("Default value")
        min_value_edit = QLineEdit(dialog)
        min_value_edit.setValidator(QIntValidator(
            -2147483648, 2147483647))
        min_value_edit.setPlaceholderText("Minimum value")
        max_value_edit = QLineEdit(dialog)
        max_value_edit.setValidator(QIntValidator(
            -2147483648, 2147483647))
        max_value_edit.setPlaceholderText("Maximum value")
        step_value_edit = QLineEdit(dialog)
        step_value_edit.setValidator(QIntValidator(
            -2147483648, 2147483647))
        step_value_edit.setPlaceholderText("Step value")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_slider_command(
            text_edit.text(), default_value_edit.text(), min_value_edit.text(), max_value_edit.text(), step_value_edit.text()))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(default_value_edit)
        layout.addWidget(min_value_edit)
        layout.addWidget(max_value_edit)
        layout.addWidget(step_value_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    def add_color_widget(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add color widget")
        text_edit = QLineEdit(dialog)
        text_edit.setPlaceholderText("Color widget name")
        default_r_edit = QLineEdit(dialog)
        default_r_edit.setValidator(QIntValidator(
            0, 255))
        default_r_edit.setPlaceholderText("Default red value")
        default_g_edit = QLineEdit(dialog)
        default_g_edit.setValidator(QIntValidator(
            0, 255))
        default_g_edit.setPlaceholderText("Default green value")
        default_b_edit = QLineEdit(dialog)
        default_b_edit.setValidator(QIntValidator(
            0, 255))
        default_b_edit.setPlaceholderText("Default blue value")
        default_a_edit = QLineEdit(dialog)
        default_a_edit.setValidator(QIntValidator(
            0, 255))
        default_a_edit.setPlaceholderText("Default alpha value")
        ok_button = QPushButton("OK", dialog)
        ok_button.clicked.connect(dialog.close)
        ok_button.clicked.connect(lambda: self.scene.add_color_widget_command(
            text_edit.text(), np.array([int(default_r_edit.text()), int(default_g_edit.text()), int(default_b_edit.text()), int(default_a_edit.text())])))
        layout = QVBoxLayout()
        layout.addWidget(text_edit)
        layout.addWidget(default_r_edit)
        layout.addWidget(default_g_edit)
        layout.addWidget(default_b_edit)
        layout.addWidget(default_a_edit)
        layout.addWidget(ok_button)
        dialog.setLayout(layout)
        dialog.exec()

    @pyqtSlot(str, np.ndarray)
    def add_color_widget_to_editor(self, name: str, default_value: np.ndarray):
        default_value = np.array(default_value)
        label = QLabel(text=name)
        color_widget = ColorWidget()
        color_widget.setCurrentColor(QColor(
            default_value[0], default_value[1], default_value[2], default_value[3]))
        self.controls[name] = color_widget
        setattr(self.scene, name, color_widget.color_tracker)
        self.controls_layout.addWidget(label)
        self.controls_layout.addWidget(color_widget)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: color_widget})

    @pyqtSlot(str, str, str, str, str)
    def add_slider_to_editor(self, name: str, default_value: str, min_value: str, max_value: str, step_value: str):
        label = QLabel(text=name)
        slider = Slider()
        slider.setOrientation(Qt.Orientation.Horizontal)
        slider.setMinimum(int(min_value))
        slider.setMaximum(int(max_value))
        slider.setValue(int(default_value))
        slider.setSingleStep(int(step_value))
        self.controls[name] = slider
        setattr(self.scene, name, slider.value_tracker)
        self.controls_layout.addWidget(label)
        self.controls_layout.addWidget(slider)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: slider})

    @pyqtSlot(str, list, str)
    def add_dropdown_to_editor(self, name: str, options: list[str], default_value: str):
        label = QLabel(text=name)
        dropdown = DropdownWidget()
        dropdown.addItems(options)
        dropdown.setCurrentText(default_value)
        self.controls[name] = dropdown
        setattr(self.scene, name, dropdown.value_tracker)
        self.controls_layout.addWidget(label)
        self.controls_layout.addWidget(dropdown)
        if self.no_controls.isVisible():
            self.no_controls.hide()
        if self.manim_studio_server is not None:
            self.add_controls_to_clients(
                self.manim_studio_server.clients, {name: dropdown})

    def save_snippet(self):
        self.communicate.save_snippet.emit(self.code_cell_edit.toPlainText())

    def save_snippet_command(self, code: str):
        controls_dialog = ControlDialog(self.controls)
        add_controls, controls_toggle = controls_dialog.exec_and_return()
        file_ = QFileDialog.getSaveFileName(
            self, "Save snippet", ".", "Manim Studio Snippet (*.mss)")
        if file_[0]:
            self.ready_to_save = True
            with open(file_[0], "w", encoding="utf-8") as f:
                f.write(code)
            if add_controls is True:
                controls = {k: v for k, v in self.controls.items()
                            if k in controls_toggle.keys() and controls_toggle[k] is True}

                def get_tup(v):
                    if isinstance(v, Slider):
                        return ("Slider", v.value(), v.minimum(), v.maximum(), v.singleStep())
                    elif isinstance(v, ColorWidget):
                        return ("ColorWidget", list(v.currentColor().getRgb()))
                    elif isinstance(v, DropdownWidget):
                        return ("DropdownWidget", [v.itemText(i) for i in range(v.count())], v.currentText())
                    elif isinstance(v, LineEditorWidget):
                        return ("LineEditorWidget", v.text())
                    elif isinstance(v, TextEditorWidget):
                        return ("TextEditorWidget", v.toPlainText())
                    elif isinstance(v, CheckboxWidget):
                        return ("CheckboxWidget", v.isChecked())

                controls = {k: get_tup(v) for k, v in controls.items()}
                with open(f"{file_[0]}.controls", "wb") as f:
                    pickle.dump(controls, f)
            else:
                if Path(f"{file_[0]}.controls").exists():
                    Path(f"{file_[0]}.controls").unlink()

    def end_scene_saving(self):
        codes = "\n".join(self.scene.codes)
        self.ready_to_save = False
        self.save_snippet_command(codes)
        if self.ready_to_save is True:
            self.end_scene()
        else:
            self.ready_to_save = True
            alert = QMessageBox(
                text="You have not saved the snippet yet.")
            alert.setWindowTitle("Snippet not saved")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()

    def save_snippet_and_run(self):
        self.save_snippet()
        self.send_code()

    def open_snippet(self):
        file_ = QFileDialog.getOpenFileName(
            self, "Open snippet", ".", "Manim Studio Snippet (*.mss)")
        if file_[0]:
            with open(file_[0], "r", encoding="utf-8") as f:
                self.code_cell_edit.setText(
                    f"{self.code_cell_edit.toPlainText()}\n{f.read()}")
            if Path(f"{file_[0]}.controls").exists():
                with open(f"{file_[0]}.controls", "rb") as f:
                    controls = pickle.load(f)
                for name, control in controls.items():
                    if name in self.controls.keys():
                        continue
                    if control[0] == "Slider":
                        self.add_slider_to_editor(
                            name, control[1], control[2], control[3], control[4])
                    elif control[0] == "ColorWidget":
                        self.add_color_widget_to_editor(
                            name, np.array(control[1]))
                    elif control[0] == "DropdownWidget":
                        self.add_dropdown_to_editor(
                            name, control[1], control[2])
                    elif control[0] == "LineEditorWidget":
                        self.add_line_editor_widget_to_editor(
                            name, control[1])
                    elif control[0] == "TextEditorWidget":
                        self.add_text_editor_to_editor(
                            name, control[1])
                    elif control[0] == "CheckboxWidget":
                        self.add_checkbox_to_editor(
                            name, control[1])
        else:
            alert = QMessageBox(
                text="No file selected.")
            alert.setWindowTitle("No file selected")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()

    def open_snippet_and_run(self):
        self.open_snippet()
        self.send_code()

    def next_slide(self):
        if self.scene.freeze is False:
            alert = QMessageBox(
                text="The scene is not paused.")
            alert.setWindowTitle("Scene not paused")
            alert.setIcon(QMessageBox.Icon.Information)
            alert.setStandardButtons(QMessageBox.StandardButton.Ok)
            alert.exec()
            return
        self.scene.freeze = False

    def end_scene(self):
        self.communicate.end_scene.emit()



import socket
from PyQt6.QtCore import QThread

from manim_studio.communicate import Communicate


class Server(QThread):
    def __init__(self, host: str, port: int, password: str, communicate: Communicate, editor):
        super().__init__()
        self.host = host
        self.port = port
        self.password = password
        self.users: dict[str, socket.socket] = {}
        self.communicate = communicate
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.bind((self.host, self.port))
        self.s.listen()
        self.clients: list[socket.socket] = []
        self.threads: list[QThread] = []
        self.running = True
        self.editor = editor

    def run(self):
        while self.running:
            try:
                client, address = self.s.accept()
            except OSError:
                break
            thread = ClientThread(
                client, address, self.password, self.communicate, self.editor, self)
            thread.start()

    def disconnect_user(self, username: str):
        self.users[username].sendall(b"exit")
        del self.users[username]

    def close(self):
        self.running = False
        for client in self.clients:
            client.sendall(b"exit")
            client.close()
        for thread in self.threads:
            thread.quit()
            thread.wait()
        self.s.close()


class ClientThread(QThread):
    def __init__(self, client: socket.socket, address: tuple[str, int], password: str, communicate: Communicate, editor, main_thread: Server):
        super().__init__()
        self.client = client
        self.address = address
        self.password = password
        self.communicate = communicate
        self.editor = editor
        self.main_thread = main_thread

    def run(self):
        msg = self.guaranteed_recv()
        try:
            msg.decode("utf-8")
        except UnicodeDecodeError:
            self.client.sendall(b"Wrong password (UnicodeDecodeError)")
            self.client.close()
            return
        if msg.decode("utf-8") == self.password:
            self.main_thread.clients.append(self.client)
            self.client.sendall(b"Correct password")
            while True:
                msg = self.guaranteed_recv()
                if msg.decode("utf-8").startswith("username="):
                    user = msg.decode("utf-8").split("=")[1]
                    if user not in self.main_thread.users:
                        self.main_thread.users[user] = self.client
                        self.client.sendall(b"Username accepted")
                    else:
                        self.client.sendall(b"Username already taken")
                        continue
                    self.communicate.show_in_status_bar.emit(
                        f"{user} connected")
                    continue
                elif msg.decode("utf-8") == "get_controls":
                    self.communicate.add_controls_to_client.emit(
                        self.client, self.editor.controls)
                    continue
                elif msg.decode("utf-8") == "exit":
                    self.main_thread.clients.remove(self.client)
                    break
                elif msg.decode("utf-8").startswith("set_slider_value"):
                    _, name, value = msg.decode("utf-8").split(" ")
                    self.communicate.set_control_value.emit(
                        name, value)
                    continue
                elif msg.decode("utf-8").startswith("set_color"):
                    _, name, r, g, b, a = msg.decode("utf-8").split(" ")
                    self.communicate.set_control_value.emit(
                        name, f"{r},{g},{b},{a}")
                    continue
                elif msg.decode("utf-8").startswith("set_dropdown"):
                    _, name, text = msg.decode("utf-8").split(" ")
                    self.communicate.set_control_value.emit(
                        name, text)
                    continue
                elif msg.decode("utf-8").startswith("set_line_edit"):
                    _, name, *words = msg.decode("utf-8").split(" ")
                    self.communicate.set_control_value.emit(
                        name, " ".join(words))
                    continue
                elif msg.decode("utf-8").startswith("set_text_edit"):
                    _, name, *words = msg.decode("utf-8").split(" ")
                    self.communicate.set_control_value.emit(
                        name, " ".join(words))
                    continue
                elif msg.decode("utf-8").startswith("set_checkbox"):
                    _, name, value = msg.decode("utf-8").split(" ")
                    self.communicate.set_control_value.emit(
                        name, value)
                    continue
                elif msg.decode("utf-8").startswith("button_clicked"):
                    _, name = msg.decode("utf-8").split(" ")
                    self.communicate.press_button.emit(name)
                    continue
                elif msg.decode("utf-8").startswith("set_position_control_value_x"):
                    _, name, value = msg.decode("utf-8").split(" ")
                    self.communicate.set_position_control_x.emit(
                        name, value)
                    continue
                elif msg.decode("utf-8").startswith("set_position_control_value_y"):
                    _, name, value = msg.decode("utf-8").split(" ")
                    self.communicate.set_position_control_y.emit(
                        name, value)
                    continue
                elif msg.decode("utf-8").startswith("set_position_control_value_z"):
                    _, name, value = msg.decode("utf-8").split(" ")
                    self.communicate.set_position_control_z.emit(
                        name, value)
                    continue
                elif msg.decode("utf-8").startswith("set_position_control_display_dot"):
                    _, name, value = msg.decode("utf-8").split(" ")
                    self.communicate.set_position_control_display_dot.emit(
                        name, value)
                    continue
                else:
                    self.communicate.update_scene.emit(msg.decode("utf-8"))
        else:
            self.client.sendall(b"Wrong password")
            self.client.close()

    def guaranteed_recv(self):
        return self.client.recv(1024)



try:
    import openai
except ImportError:
    AIWidget = None
else:
    from PyQt6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QTextEdit, QDialog, QLabel, QLineEdit, QComboBox
    from PyQt6.QtGui import QIntValidator
    from .communicate import Communicate

    SYSTEM_INSTRUCTIONS = "Generate a Python code snippet to generate a Manim animation according to the prompt. " \
        + "The response must contain only one snippet. " \
        + "Never explain the code that you are generating. " \
        + "The code must be valid Python code. " \
        + "The code must not contain any comments. " \
        + "The code must not contain any imports unless it is necessary. " \
        + "Assume you already import Manim and you defined the construct() method of a scene. That means " \
        + "that you won't indent nor write 'from manim import *', nor 'class <SceneName>(Scene):', nor 'def construct(self):'. " \
        + "Your answer must be a code block, it starts with ```python and ends with ```."

    class AIWidget(QWidget):
        def __init__(self, communicate: Communicate, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setWindowTitle("Manim Studio - AI Power")
            self.openai_api_key_is_set = False
            self.communicate = communicate
            self.layout_ = QVBoxLayout()
            self.send_code_button = QPushButton("Send code")
            self.send_code_button.clicked.connect(self.send_code)
            self.response_edit = QTextEdit()
            self.response_edit.setPlaceholderText("Response")
            self.response_edit.setReadOnly(True)
            self.layout_.addWidget(self.response_edit)
            self.prompt_edit = QTextEdit()
            self.prompt_edit.setPlaceholderText("Enter the prompt")
            self.layout_.addWidget(self.prompt_edit)
            self.tokens_edit = QLineEdit()
            self.tokens_edit.setValidator(QIntValidator())
            self.tokens_edit.setPlaceholderText(
                "Enter the max tokens (default: 100)")
            self.layout_.addWidget(self.tokens_edit)
            self.send_prompt_button = QPushButton("Send prompt")
            self.send_prompt_button.clicked.connect(self.send_prompt)
            self.select_gpt_version_dropdown = QComboBox()
            self.gpt_version = None
            self.select_gpt_version_dropdown.setPlaceholderText(
                "Select GPT version")
            gpt_dict = {
                "GPT 3.5 Turbo": "gpt-3.5-turbo",
                "GPT 4": "gpt-4"
            }
            self.select_gpt_version_dropdown.addItems(
                ["GPT 3.5 Turbo", "GPT 4"])
            self.select_gpt_version_dropdown.currentIndexChanged.connect(
                lambda: setattr(self, "gpt_version", gpt_dict[self.select_gpt_version_dropdown.currentText()]))
            self.layout_.addWidget(self.send_prompt_button)
            self.layout_.addWidget(self.select_gpt_version_dropdown)
            self.layout_.addWidget(self.send_code_button)
            self.setLayout(self.layout_)
            self.messages = [{"role": "system",
                              "content": SYSTEM_INSTRUCTIONS}]

        def send_prompt(self):
            if self.openai_api_key_is_set is False:
                dialog = QDialog()
                self.api_key_dialog = dialog
                dialog.setWindowTitle("OpenAI API key")
                dialog.layout_ = QVBoxLayout()
                dialog.api_key_edit = QTextEdit()
                dialog.api_key_edit.setPlaceholderText(
                    "Enter your OpenAI API key")
                dialog.layout_.addWidget(dialog.api_key_edit)
                dialog.warn_label = QLabel(
                    text="Please set your OpenAI API key. Once done, send the prompt again.")
                dialog.layout_.addWidget(dialog.warn_label)
                dialog.ok_button = QPushButton("OK")
                dialog.ok_button.clicked.connect(
                    lambda: self.set_api_key(dialog.api_key_edit.toPlainText()))
                dialog.layout_.addWidget(dialog.ok_button)
                dialog.setLayout(dialog.layout_)
                dialog.exec()
                return
            if self.gpt_version is None:
                dialog = QDialog()
                dialog.setWindowTitle("Select GPT version")
                dialog.layout_ = QVBoxLayout()
                dialog.warn_label = QLabel(
                    text="Please select a GPT version. Once done, send the prompt again.")
                dialog.layout_.addWidget(dialog.warn_label)
                dialog.ok_button = QPushButton("OK")
                dialog.ok_button.clicked.connect(
                    lambda: dialog.close())
                dialog.layout_.addWidget(dialog.ok_button)
                dialog.setLayout(dialog.layout_)
                dialog.exec()
                return
            try:
                self.messages.append({"role": "user",
                                      "content": self.prompt_edit.toPlainText()})
                code_completion = openai.ChatCompletion.create(
                    model=self.gpt_version,
                    messages=self.messages,
                    max_tokens=int(self.tokens_edit.text() or 100),
                )
            except openai.OpenAIError as e:
                self.openai_api_key_is_set = False
                self.communicate.alert.emit(e)
            else:
                response = code_completion.choices[0].message.content
                if response.startswith("```python") and response.endswith("```"):
                    response = response[len("```python"): -len("```")]
                self.response_edit.setReadOnly(False)
                self.response_edit.setText(response)

        def send_code(self):
            self.communicate.update_scene.emit(
                self.response_edit.toPlainText())
            self.response_edit.setReadOnly(True)

        def set_api_key(self, api_key: str):
            openai.api_key = api_key
            self.openai_api_key_is_set = True
            self.api_key_dialog.close()



from manim import ValueTracker, Mobject
from colour import Color
import numpy as np


class ColorValueTracker(ValueTracker):
    def __init__(self, value=np.array([0, 0, 0, 1]), **kwargs):
        Mobject.__init__(self, **kwargs)
        self.set_points(np.zeros((1, 4)))
        self.set_value(value)

    def get_value(self) -> Color:
        return Color(rgb=np.array(self.points[0, :3])).hex, self.points[0, 3]

    def set_value(self, value: np.ndarray):
        self.points[0, :4] = np.array(value)
        return self



import socket
from PyQt6.QtWidgets import QWidget, QLabel, QVBoxLayout, QPushButton, QTextEdit, QMessageBox, QStatusBar
from .control_dialog import ClientControls


class ManimStudioClient(QWidget):
    def __init__(self, host: str, port: int, username: str, password: str):
        super().__init__()
        self.setWindowTitle("Manim Studio Client")
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.connect((self.host, self.port))
        self.s.sendall(self.password.encode("utf-8"))
        msg = self.s.recv(1024)
        if msg.decode("utf-8") == "Correct password":
            self.s.sendall(f"username={self.username}".encode("utf-8"))
            answer = self.s.recv(1024)
            if answer.decode("utf-8") == "Username already taken":
                QMessageBox.critical(
                    self, "Error", "Username already taken")
                self.close()
                self.success = False
                return
            else:
                self.success = True
            self.setLayout(QVBoxLayout())
            self.label = QLabel()
            self.label.setText(
                f"Connected to {self.host}:{self.port}")
            self.layout().addWidget(self.label)
            self.code_edit = QTextEdit()
            self.code_edit.setPlaceholderText("Enter the code")
            self.layout().addWidget(self.code_edit)
            self.send_button = QPushButton("Send (Ctrl+Return)")
            self.send_button.setShortcut("Ctrl+Return")
            self.send_button.clicked.connect(self.send_code)
            self.layout().addWidget(self.send_button)
            self.status_bar = QStatusBar()
            self.layout().addWidget(self.status_bar)
            self.controls_dialog = ClientControls(self, self.s)
        else:
            QMessageBox.critical(
                self, "Error", msg.decode("utf-8"))
            self.close()

    def closeEvent(self, event):
        self.s.sendall(b"exit")
        self.s.close()
        super().closeEvent(event)

    def send_code(self):
        code = self.code_edit.toPlainText()
        self.s.sendall(code.encode("utf-8"))
        self.code_edit.clear()
        self.status_bar.showMessage("Code sent")
